from . import JINJA_ENV


def page(request, template_name, context, _env=JINJA_ENV):
    template = _env.get_template(template_name)
    html = template.render(**context)
    if isinstance(html, unicode):
        return html.encode('utf-8')
    return html



from logging import getLogger
import os

logger = getLogger(__name__)

try:
    from django.conf import settings
    has_django = True
except (ImportError, Exception), e:
    logger.warning(str(e))
    has_django = False

import json


if has_django:
    INTERVAL = getattr(settings, 'TIEMPO_INTERVAL', 5)
    THREAD_COUNT = getattr(settings, 'TIEMPO_THREADS', 1)
    TASK_GROUPS = ['ALL'] + getattr(settings, 'TIEMPO_GROUPS', [])
    RESULT_LIFESPAN = getattr(settings, 'TIEMPO_RESULT_LIFESPAN_DAYS', 1)
    DEBUG = settings.DEBUG

    REDIS_HOST = settings.REDIS_HOST
    REDIS_PORT = settings.REDIS_PORT
    REDIS_QUEUE_DB = settings.REDIS_QUEUE_DB
    REDIS_PW = settings.REDIS_PW

else:
    INTERVAL = os.environ.get('TIEMPO_INTERVAL', 5)
    THREAD_COUNT = os.environ.get('TIEMPO_THREADS', 1)
    TASK_GROUPS = ['ALL'] + os.environ.get('TIEMPO_GROUPS', [])
    RESULT_LIFESPAN = os.environ.get('TIEMPO_RESULT_LIFESPAN_DAYS', 1)
    DEBUG = os.environ.get('TIEMPO_DEBUG', False)

    REDIS_HOST = os.environ.get('TIEMPO_REDIS_HOST', 'localhost')
    REDIS_PORT = os.environ.get('TIEMPO_REDIS_PORT', 6379)
    REDIS_QUEUE_DB = os.environ.get('TIEMPO_REDIS_QUEUE_DB', 12)
    REDIS_PW = os.environ.get('TIEMPO_REDIS_PW', None)

TASK_PATHS = json.loads(os.environ.get('TIEMPO_TASK_PATHS', '["tiempo.demo"]'))



from .contrib.django.utils import six

from tiempo.conf import TASK_PATHS

import chalk
import importlib
import os


def import_tasks():
    for app in TASK_PATHS:
        if not app.split('.')[-1] == 'tasks':
            module = importlib.import_module(app)
            filename = os.path.join(module.__path__[0], 'tasks.py')
            if not os.path.isfile(filename):
                chalk.yellow(app + ' does not have a tasks module.')
                continue
            else:
                app = app + '.tasks'

        chalk.blue(app.upper() + ': imported tasks from %s' % app)
        importlib.import_module(app)


# /////////////////////////////////////////////////////////////////////////////
# Django Rubbish
class Promise(object):
    """
    This is just a base class for the proxy class created in
    the closure of the lazy function. It can be used to recognize
    promises in code.
    """
    pass


def force_bytes(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to smart_bytes, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    if isinstance(s, six.memoryview):
        s = bytes(s)
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    if strings_only and (s is None or isinstance(s, int)):
        return s
    if isinstance(s, Promise):
        return six.text_type(s).encode(encoding, errors)
    if not isinstance(s, six.string_types):
        try:
            if six.PY3:
                return six.text_type(s).encode(encoding)
            else:
                return bytes(s)
        except UnicodeEncodeError:
            if isinstance(s, Exception):
                # An Exception subclass containing non-ASCII data that doesn't
                # know how to print itself properly. We shouldn't raise a
                # further exception.
                return b' '.join(
                    [
                        force_bytes(arg, encoding, strings_only, errors)
                        for arg in s
                    ]
                )
            return six.text_type(s).encode(encoding, errors)
    else:
        return s.encode(encoding, errors)



from setuptools import setup, find_packages
import os
import tiempo


def file_name(rel_path):
    dir_path = os.path.dirname(__file__)
    return os.path.join(dir_path, rel_path)


def readlines(rel_path):
    with open(file_name(rel_path)) as f:
        ret = f.readlines()
    return ret


setup(
    author="Damon",
    name="tiempo",
    version=tiempo.__version__,
    packages=find_packages(exclude=["tests*", ]),
    url="https://github.com/hangarunderground/tiempo",
    description="Twisted task scheduling for django",
    scripts=['tiempo/scripts/metronome', ],
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    keywords=['twisted', 'tasks', 'redis', 'scheduling'],
    install_requires=readlines('requirements.txt'),
    extras_require={'dev': ['ipdb', ], 'django': ['django', ]},
    include_package_data=True
)




from . import TIEMPO_REGISTRY, RECENT_KEY
from .task import Task
from .conn import REDIS
from .conf import INTERVAL, THREAD_COUNT, TASK_GROUPS, RESULT_LIFESPAN, DEBUG
from .exceptions import ResponseError

from twisted.internet import threads, task
from dateutil.relativedelta import relativedelta
from cStringIO import StringIO
from logging import getLogger

import datetime
import sys
import chalk
import json
import pytz
import traceback


logger = getLogger(__name__)


def utc_now():
    return datetime.datetime.now(pytz.utc)


class CaptureStdOut(list):

    def __init__(self, *args, **kwargs):

        self.task = kwargs.pop('task')
        self.start_time = utc_now()

        super(CaptureStdOut, self).__init__(*args, **kwargs)

    def __enter__(self):
        if DEBUG:
            pass
            # return self
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        if DEBUG:
            # return
            pass

        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout

    def finished(self, timestamp=None):
        if DEBUG:
            pass
            # return

        self.timestamp = timestamp
        if not timestamp:
            self.timestamp = utc_now()

        task_key = '%s:%s' % (self.task.key, self.task.uid)
        expire_time = int(((self.timestamp + relativedelta(
            days=RESULT_LIFESPAN)) - self.timestamp).total_seconds())

        pipe = REDIS.pipeline()
        pipe.zadd(RECENT_KEY, self.start_time.strftime('%s'), task_key)
        pipe.set(self.task.uid, self.format_output())
        pipe.expire(self.task.uid, expire_time)
        pipe.execute()

    def format_output(self):

        return json.dumps(
            {
                'task': self.task.key,
                'uid': self.task.uid,
                'start_time': self.start_time.isoformat(),
                'end_time': self.timestamp.isoformat(),
                'duration': (self.timestamp - self.start_time).total_seconds(),
                'output': self
            }
        )


def get_task_keys(TASK_GROUPS):
    """
        creates a dictionary containing a mapping every combination of
        cron like keys with the corresponding amount of time
        until the next execution of such a task if it were to be executed
        now.


        ie:

        {'*.*.*': <1 minute from now>,}
        {'*.5.*': <1 minute from now>,}
        {'*.5.11': <24 hours from now>,}

    """

    time_keys = {}

    now = utc_now()

    # first any tasks that want to happen this minute.
    for tg in TASK_GROUPS:

        # every day, every hour, every minute
        expires = now + datetime.timedelta(minutes=1)
        time_keys['*.*.*'] = expires

        # every day, every hour, this minute
        expires = now + datetime.timedelta(hours=1)
        time_keys[now.strftime('*.*.%M')] = expires

        # every day, this hour, this minute
        expires = now + datetime.timedelta(days=1)
        time_keys[now.strftime('*.%H.%M')] = expires

        # this day, this hour, this minute
        expires = now + relativedelta(months=1)
        time_keys[now.strftime('%d.%H.%M')] = expires

        # this day, this hour, every minute
        expires = now + datetime.timedelta(minutes=1)
        time_keys[now.strftime('%d.%H.*')] = expires

        # this day, every hour, every minute
        expires = now + datetime.timedelta(minutes=1)
        time_keys[now.strftime('%d.*.*')] = expires

        # every day, this hour, every minute
        expires = now + datetime.timedelta(minutes=1)
        time_keys[now.strftime('*.%H.*')] = expires

    # logger.debug(time_keys.keys())
    return time_keys


class ThreadManager(object):

    def __init__(self, number):
        self.active_task = None
        self.number = number

    def __repr__(self):
        return 'tiempo thread %d' % self.number

    def control(self):
        chalk.blue('%r checking for work' % self)

        """
            First we check our task registry and check if there are any
            tasks that should be running right this second.

            If there are, we queue them.

            Next we execute queued tasks

        """

        now = utc_now()
        tomorrow = now + datetime.timedelta(days=1)

        time_tasks = get_task_keys(TASK_GROUPS)

        _tasks = [
            t for t in TIEMPO_REGISTRY.values()
            if t.group in TASK_GROUPS and t.periodic
        ]
        for _task in _tasks:
            stop_key = '%s:schedule:%s:stop' % (_task.group, _task.key)

            expire_key = time_tasks.get(_task.get_schedule())
            # the expire key is a datetime which signifies the NEXT time this
            # task would run if it was to be run right now.
            #
            # if there is no expire key it means that there are no tasks that
            # should run right now.

            if expire_key:
                # if we are here, it means we have a task that could run
                # right now if no other worker has already started it.
                try:

                    # create a key whose existence prevents this task
                    # from executing.  This will ensure that it only runs
                    # once across the whole network.

                    if REDIS.setnx(stop_key, 0):
                        logger.debug(
                            'running task %r because: %r',
                            _task, _task.get_schedule()
                        )
                        logger.debug('expire key: %r', expire_key)
                        logger.debug('now: %r', now)
                        logger.debug('stop key: %r', stop_key)
                        logger.debug(
                            'seconds: %r', (expire_key - now).total_seconds()
                        )
                        # the creation of the key was successful.
                        # this means that previous keys whose existence
                        # postponed execution for this task have expired.

                        # We will now set the expiration on this key to expire
                        # at the time of the next appropriate running of this
                        # task

                        REDIS.expire(
                            stop_key,
                            int(float((expire_key - now).total_seconds()))
                        )

                        # queue it up
                        _task.soon()

                        logger.debug('next will be: %r', expire_key)

                except ResponseError as e:
                    print e
                except BaseException as e:
                    print e

        try:
            for g in TASK_GROUPS:
                if not self.active_task:
                    task = REDIS.lpop(g)
                    if task:
                        logger.debug(
                            'RUNNING TASK on thread %r: %s' % (self, task)
                        )
                        self.active_task = True
                        threads.deferToThread(run_task, task, self)

        except BaseException as e:
            chalk.red('%s got an error' % self)
            print traceback.format_exc()

    def cleanup(self):
        pass
        # print 'thread %d cleaning up' % thread

    def start(self):
        task.LoopingCall(self.control).start(INTERVAL)


def run_task(task, thread):

    task = Task.rehydrate(task)
    chalk.yellow('%r: %s' % (thread, task.key))
    with CaptureStdOut(task=task) as output:
        try:
            task.start()
            task.run()
            task.finish()
        except:
            print traceback.format_exc()
            task.finish(traceback.format_exc())

    output.finished()
    thread.active_task = False


def thread_init():
    if THREAD_COUNT:
        chalk.green('current time: %r' % utc_now())
        chalk.green(
            'Found %d thread(s) specified by TIEMPO_THREADS' % THREAD_COUNT
        )

        for i in range(0, THREAD_COUNT):
            tm = ThreadManager(i)
            tm.start()
